#ifndef __hcd_h__
#define __hcd_h__

#include<stdio.h>
#include<stdarg.h>
#include <stdlib.h>
#include<sys/socket.h>
#include<sys/types.h>
#include<netinet/in.h>
#include<time.h>
#include<errno.h>

#define BUFFLEN 8191
#define CHHSHSIZE 32
#define MSGMAX 4
#define DEFAULT_PORT    1234
#define DEFAULT_LISTCNT 5

typedef struct t_hcd_buff {
	char buff [BUFFLEN+1];
	int len;
	int send_start;
	int hdstat;
	int hdend;
	int explen;
} t_hcd_buff;

int hcd_buff_append (t_hcd_buff *hb, const char *buff, int bsize);
int hcd_buff__printf (t_hcd_buff *buff, const char *fmt, ...);
int hcd_buff__append_esc (t_hcd_buff *buff, const char *s);

typedef struct t_hcd_client {
	t_hcd_buff b_in;
	t_hcd_buff b_out;
	int fd;
	int type;
	struct t_hcd_request *req;
	struct t_hcd_chat *chat;
	struct t_hcd_client *cl_nxt;
	struct t_hcd_client *cl_prev;
	struct t_hcd_user *user;
} t_hcd_client;

typedef struct t_hcd_user {
	char hsh [CHHSHSIZE+1];
	char *title;
	struct t_hcd_user *next;
} t_hcd_user;

t_hcd_user *hcd_user__new (struct t_hcd_chat *chat);
t_hcd_user *hcd_chat__user_find (struct t_hcd_chat *chat, const char *hsh);


enum {
	REQ_UNKNOWN,
	REQ_CHAT_CREATE,
	REQ_CHAT_GET,
	REQ_MSG_POST,
	REQ_MSG_GET,
	REQ_STATIC_GET
};

enum {
	F_CHAT_PUBLIC = 1
};

typedef struct t_hcd_request {
	int type;
	char *path;
	int id;
	int flags;
	void *data;
} t_hcd_request;

void hcd_request__free (t_hcd_request *req);
void hcd_client_free (t_hcd_client *cl);
t_hcd_client *hcd_client_new (int fd);

enum {
	CHAT_PUBLIC = 1
};

typedef struct t_hcd_msg {
	int id;
	t_hcd_client *cli;
	struct t_hcd_user *user;
	time_t tim;
	char *msg;
} t_hcd_msg;

typedef struct t_hcd_msqu {
	int max;
	int min;
	t_hcd_msg msgs [MSGMAX];
} t_hcd_msqu;

t_hcd_msqu *hcd_msqu__new ();
t_hcd_msg *hcd_msqu__find (t_hcd_msqu *q, int id);
void hcd_msqu__free (t_hcd_msqu *q);
t_hcd_msg *hcd_msqu__add_msg (t_hcd_msqu *q, t_hcd_client *cli, const char *msg);

typedef struct t_hcd_chat {
	char hsh [CHHSHSIZE+1];
	char *title;
	int flags;
	struct t_hcd_msqu *msqu;
	struct t_hcd_chat *next;
	struct t_hcd_client *clients;
	struct t_hcd_user *users;
} t_hcd_chat;

void hcd_chat__attach_client (t_hcd_chat *chat, t_hcd_client *cl);

typedef struct t_hcd_ctx {
	int port;
	int listcnt;
	struct sockaddr_in sr_addr;
	int sock;
	struct pollfd *poll_arr;
	t_hcd_client **cli_arr;
	int poll_max;
	int poll_cnt;
	int interval;
	t_hcd_chat *chats;
} t_hcd_ctx;


t_hcd_request *hcd_client_parse_header (t_hcd_ctx *ctx, t_hcd_client *cl);
t_hcd_chat *hcd_chat__new (t_hcd_ctx *ctx);

t_hcd_chat *hcd_ctx__chat_find (t_hcd_ctx *ctx, const char *hsh);

typedef struct t_hcd_url_param {
	char *name;
	char *value;
	struct t_hcd_url_param *next;
} t_hcd_url_param;

typedef struct t_hcd_url {
	t_hcd_url_param *params;
	char *path;
} t_hcd_url;

void hcd_url__free (t_hcd_url *);
t_hcd_url *hcd_url__parse (const char*);

#define NEW(x) (x *)memset(malloc(sizeof(x)), 0, sizeof(x))
enum {
	ERR =   1,
	TRA =   2,
	MSG =   4,
	SYS =   8,
	DTA =  16,
	TRA2=  32
};

void hcd_set_log_file (FILE *f);
void hcd_log (int level, const char *fmt, ...);

#define L(typ,fmt,args...) hcd_log(typ,__FILE__":%s:%d: " fmt, __func__, __LINE__, ##args)



#endif
