#include <hcd.h>
#include <hcd_reqparse.h>
#include <string.h>

typedef enum {
	_end = 0, start = 1, end, geG, geGE, geGET, geGET_, geP, gePO, gePOS, gePOST,
	gePOST_, gePATH, gePATH_PARNAM, gePATH_PARVAL, gePATH_CR, gePATH_CRLF, gePATH_CRLFCR, geHEADER, geHEADER_VAL, geHEADER_CR,
	geHEADER_CRLF, geHEADER_CRLFCR, gePROTO, geHEADER_NAM
} t_states;

typedef enum {
	METHOD_GET = 1, METHOD_POST = 2
} t_methods;

typedef enum {
	PARSED_OK = 0, PARSED_OUT, PARSED_FAILED_FUNC, PARSED_FAILURE, PARSED_NOMATCH, PARSED_OVERFLOW
} t_results ;

typedef enum {
	S_FAILURE = 2, S_FINISH = 4, S_APPEND = 8, S_CLEAR = 16, S_STORE = 32
}t_flags ;

typedef struct t_state {
	t_states from;
	const char *match;
	t_states to;
	int flag;
	int (*fce)(struct t_parsectx *, const struct t_state *s);
	int param;
} t_state;

typedef struct t_parsectx {
	const t_state *states;
	t_states st;
	char buff [4096];
	char buff2 [4096];
	unsigned int buffpos;
	void *data;
	t_hcd_request *req;
	t_hcd_ctx *hctx;
	t_hcd_client *cl;
} t_parsectx;

static int set_path (t_parsectx *ctx, const t_state *s) 
{
	char *p;

	if (! strncmp (ctx->buff, "/_", 2) || ! strcmp (ctx->buff, "/")) {
		ctx->req->type = REQ_STATIC_GET;
		ctx->req->path = NULL;
	} else if (! strncmp (ctx->buff, "/chat_create", 12)) {
		ctx->req->type = REQ_CHAT_CREATE;

		ctx->req->flags = 0;
	} else if (! strncmp (ctx->buff, "/msg/", 5)) {
		t_hcd_chat *chat;
		p = strchr (ctx->buff + 5, '/');

		if (p) *p = '\0';
		else {
			L (ERR, "Request of incompatible format %s", ctx->buff + 5);
			goto error;
		} 

		chat = hcd_ctx__chat_find (ctx->hctx, ctx->buff + 5);
		if (! chat) {
			L (ERR, "Request for nonexisting chat %s", ctx->buff + 5);
			goto error;
		}
		ctx->req->id = atoi (p + 1);
		ctx->cl->chat = chat;
		ctx->req->type = REQ_MSG_GET;
	} else if (! strncmp (ctx->buff, "/chat/", 6)) {
		t_hcd_chat *chat;

		chat = hcd_ctx__chat_find (ctx->hctx, ctx->buff + 6);
		if (! chat) {
			L (ERR, "Request for nonexisting chat %s", ctx->buff + 6);
			goto error;
		} 
		ctx->cl->chat = chat;
		ctx->req->type = REQ_CHAT_GET;
	} else if (! strncmp (ctx->buff, "/send/", 6)) {
		t_hcd_chat *chat;
		ctx->req->type = REQ_MSG_POST;

		chat = hcd_ctx__chat_find (ctx->hctx, ctx->buff + 6);
		if (! chat) {
			L (ERR, "Request for nonexisting chat %s", ctx->buff + 6);
			goto error;
		} 
		ctx->cl->chat = chat;

	} else {
		L (ERR, "Unknown path \'%s\'", ctx->buff);
		return (1);
	}



	return (0);
error: 
	return (1);
}

static int set_header (t_parsectx *ctx, const t_state *s) 
{
	char *n;

	if (ctx->req->type == REQ_MSG_POST) {
		if (! strcmp (ctx->buff2, "Content-Length")) {
			n = ctx->buff;
			while (*n == ' ') n ++;
			ctx->cl->b_in.explen = atoi (n);
		} else if (! strcmp (ctx->buff2, "Cookie")) {
			n = strstr(ctx->buff, "USER=");
			if (n) {
				t_hcd_user *user;
				user = hcd_chat__user_find (ctx->cl->chat, n + 5);
				ctx->cl->user = user;
			}
		}
	}

	return (0);
}

static int set_param (t_parsectx *ctx, const t_state *s) 
{
	if (ctx->req->type == REQ_CHAT_CREATE) {
		if (! strcmp (ctx->buff2, "public") && ! strcmp (ctx->buff, "on")) ctx->req->flags |= F_CHAT_PUBLIC;
		if (! strcmp (ctx->buff2, "title")) {
			ctx->req->data = strdup (ctx->buff);
		}
		
	}
	return (0);
}

static int set_method (struct t_parsectx *ctx, const struct t_state *s) 
{
	return (0);
}

void parsectx__init (t_parsectx *ctx)
{
	ctx->buffpos = 0;
	ctx->buff [0] = '\0';
}

static int match (char c, const char *m) {
	if (m [0] == '=' && strchr (m + 1, c)) return 1;
	if (m [0] == '*') return 1;
	if (m [0] == '!' && ! strchr (m + 1, c)) return 1;
	return 0;
}

static int parsectx__go (t_parsectx *ctx, char c) {
	const t_state *s;

	for (s = ctx->states; s->from != _end; s ++) {
		if (s->from != ctx->st) continue;
		if (match (c, s->match)) {
			if (s->flag & S_APPEND) { 
				ctx->buff [ctx->buffpos++] = c; 
				ctx->buff [ctx->buffpos] = '\0';
				if (ctx->buffpos >= sizeof (ctx->buff)) return (PARSED_OVERFLOW);
			}
			if (s->fce) {
				if (s->fce (ctx, s)) {
					L (ERR, "parser / function error \'%c\' (%d,%d)", c, s->from, s->to);
					return (PARSED_FAILURE);
				}
			}
			if (s->flag & S_STORE) {
				memcpy (ctx->buff2, ctx->buff, ctx->buffpos + 1);
			}
			if (s->flag & S_CLEAR) {
				ctx->buffpos = 0;
				ctx->buff [0] = '\0';
			}

			if (s->flag & S_FAILURE) return (PARSED_FAILURE);
			if (s->flag & S_FINISH) {
				ctx->st = _end;
				return (PARSED_OK);
			}
			ctx->st = s->to;
			return (PARSED_OK);
		}
	}
	L (ERR, "parser / nomatch", 0);
	return (PARSED_NOMATCH);
}

/* method to walk through finite state automaton */
int parsectx__proceed (t_parsectx *ctx, const char *text) {
	char c;
	int r;
	while (ctx->st != _end) {
		c = *text;
		text ++;

		r = parsectx__go (ctx, c);

		if (r != PARSED_OK) return (r);
		if (c == '\0') return (PARSED_OUT);

	}
	return (PARSED_OK);
}

void parsectx__free (t_parsectx *ctx)
{
	free (ctx);
}

/* finite state automaton definition */
static const t_state  lst_states []  = {
	{ start,            "=G",  geG,             0,        NULL, 0},
	{ start,            "=P",  geP,             0,        NULL},
	{ geG,              "=E",  geGE,            0,        NULL},
	{ geGE,             "=T",  geGET,           0,        NULL},
	{ geP,              "=O",  gePO,            0,        NULL},
	{ gePO,             "=S",  gePOS,           0,        NULL},
	{ gePOS,            "=T",  gePOST,          0,        NULL},
	{ gePOST,           "= ",  gePATH,          0,        set_method, METHOD_POST},
	{ geGET,            "= ",  gePATH,          0,        set_method, METHOD_GET},
	{ gePATH,           "= ",  gePROTO,         S_CLEAR,  set_path},
	{ gePATH,           "=?",  gePATH_PARNAM,   S_CLEAR,  set_path},
	{ gePATH_PARNAM,    "= ",  gePROTO,         S_CLEAR,  set_param, 1},
	{ gePATH_PARNAM,    "==",  gePATH_PARVAL,   S_STORE|S_CLEAR, NULL},
	{ gePATH_PARNAM,    "*",   gePATH_PARNAM,   S_APPEND, NULL},
	{ gePATH_PARVAL,    "= ",  gePROTO,         S_CLEAR,  set_param, 0},
	{ gePATH_PARVAL,    "=&",  gePATH_PARNAM,   S_CLEAR,  set_param, 0},
	{ gePATH_PARVAL,    "*",   gePATH_PARVAL,   S_APPEND, NULL},
	{ gePATH,           "*",   gePATH,          S_APPEND, NULL},
	{ gePROTO,          "!\r", gePROTO,         0,        NULL},
	{ gePROTO,          "=\r", gePATH_CR,       0,        NULL},
	{ gePATH_CR,        "=\n", geHEADER_CRLF,   0,        NULL},
	{ geHEADER,         "=\r", geHEADER_CR,     0,        NULL},
	{ geHEADER_CR,      "=\n", geHEADER_CRLF,   0,        NULL},
	{ geHEADER_CRLF,    "=\r", geHEADER_CRLFCR, 0,        NULL},
	{ geHEADER_CRLFCR,  "=\n", _end,            S_FINISH,  NULL},
	{ geHEADER,         "=:",  geHEADER_VAL,    S_STORE|S_CLEAR, NULL},
	{ geHEADER,         "*",   geHEADER,        S_APPEND, NULL},
	{ geHEADER_CRLF,    "*",   geHEADER,        S_APPEND, NULL},
	{ geHEADER_VAL,     "=\r", geHEADER_CR,     S_CLEAR,  set_header, 0},
	{ geHEADER_VAL,     "*",   geHEADER_VAL,    S_APPEND, NULL, 0},
	{ _end,              NULL, _end,   0, NULL }
};

t_parsectx *parsectx__new (t_hcd_ctx *hctx, t_hcd_client *client, t_hcd_request *req)
{
	t_parsectx *ctx;

	ctx = NEW (t_parsectx);
	ctx->req = req;
	ctx->hctx = hctx;
	ctx->cl = client;
	ctx->st = start;

	parsectx__init (ctx);

	ctx->states = lst_states;

	return (ctx);
}

#ifdef _TEST
int main (int argc, char **argv)
{
	const char *tst = "GET /chat_create?cc=123&vve=444\r\nCookie: USER=vee\r\nContent-length: 123\r\n\r\n";

	t_parsectx *ctx;
	t_hcd_request *req;

	hcd_set_log_file (stdout);

	req = hcd_request__new ();

	ctx = parsectx__new (NULL, NULL, req);
	if (parsectx__proceed (ctx, tst)) {
		fprintf (stderr, "Cannot proceed \n");
	}

	printf ("REQTYPE=%d\n", req->type);
	parsectx__free (ctx);
	return 0;
}
#endif
