#include<hcd.h>
#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include<signal.h>
#include<strings.h>
#include<sys/types.h>
#include<sys/socket.h>
#include<sys/stat.h>
#include<unistd.h> 
#include<fcntl.h>
#include<netdb.h>
#include<errno.h>
#include<poll.h>
#include<arpa/inet.h> 


t_hcd_ctx *ctx__new (int siz)
{
	t_hcd_ctx *c;
	int i;

	c = NEW (t_hcd_ctx);
	c->poll_arr = malloc (sizeof (struct pollfd) * siz);
	c->poll_max = siz;
	c->cli_arr = malloc (sizeof (t_hcd_client*) * siz);

	for (i = 0; i < siz; i ++) {
		c->poll_arr [i].fd = -1;
		c->cli_arr [i] = NULL;
	}

	return (c);
}

/**
 * Bind ctx to specified port
 */
int ctx__bind (t_hcd_ctx *ctx)
{
	int sockopt;
	struct addrinfo *n, *no, hostinf;
	char sport [128];
	int r, sock;

	snprintf (sport, sizeof (sport), "%d", ctx->port);

	memset (& hostinf, 0, sizeof (hostinf));

	hostinf.ai_family = AF_UNSPEC;
	hostinf.ai_socktype = SOCK_STREAM;
	hostinf.ai_flags = AI_PASSIVE;

	r = getaddrinfo (NULL, sport, &hostinf, &n);
	
	if (r || ! n) {
		fprintf (stderr, "Cannot getaddrinfo socket %d\n", r);
		return (-2);
	}

	no = n;

	/* go through interfaces, try to create socket and bind it*/
	for (;n; n = n->ai_next) {
		sock = socket (n->ai_family, n->ai_socktype, n->ai_protocol);

		if (sock == -1) {
			fprintf (stderr, "Cannot create socket.. %d (%s)", errno, strerror (errno));
			freeaddrinfo (no);
			return (-2);
		}

		sockopt = 1;
		
		if (setsockopt (sock, SOL_SOCKET, SO_REUSEADDR, (void*) &sockopt, sizeof (sockopt)) != 0){
			fprintf (stderr, "error setting socket options (%s).", strerror (errno));
			return (-1);
		}

		if (bind (sock, n->ai_addr, n->ai_addrlen) == 0) break;
		
		close (sock);

		sock = -1;
	}

	if (sock == -1) {
		fprintf (stderr, "Cannot create socket.\n");
		return (-3);
	}

	ctx->sock = sock;
	memset (& ctx->sr_addr, 0, sizeof (struct sockaddr_in));

	ctx->sr_addr.sin_family = n->ai_family;
/*	ctx->sr_addr.sin_addr.s_addr = n->ai_addr;*/
	ctx->sr_addr.sin_port = htons (ctx->port);


	listen (ctx->sock, 5);

	ctx->poll_arr [0].fd = ctx->sock;
	/*FD_ZERO (&(ctx->poll_arr [0].events));
	FD_SET (0, & ctx->poll_arr [0].events);*/
	ctx->poll_arr [0].events = POLLIN;
	ctx->poll_cnt = 1;
	return (0);
}

/**
 * accept() connection on socket of specified index.
 * Put newly created socket on the end of the array.
 */
int ctx__accept (t_hcd_ctx *ctx, int id)
{
        struct sockaddr_in cl_addr;
	socklen_t cl_len;
	int newfd, r;
	cl_len = sizeof (struct sockaddr_in);

	newfd = accept (ctx->poll_arr [id].fd, (struct sockaddr*) & cl_addr, (socklen_t*) &cl_len);

	if (newfd == -1) {
		L (ERR|SYS, "Cannot accept! %s",strerror (errno));
		return (-2);
	}

	r = fcntl (newfd, F_GETFL);
	fcntl (newfd, F_SETFL, r | O_NONBLOCK);

	ctx->poll_arr [ctx->poll_cnt].fd = newfd;
	ctx->poll_arr [ctx->poll_cnt].events = POLLIN | POLLNVAL;
	ctx->cli_arr [ctx->poll_cnt] = hcd_client_new (newfd);

	ctx->poll_cnt ++;

	return (0);
}

/**
 * shutdown() and close() specified socket.
 * Move the array to fill empty slot.
 */
int ctx__shutdown (t_hcd_ctx *ctx, int i)
{
	if (shutdown (ctx->poll_arr [i].fd, SHUT_RDWR)) {
		L (ERR|SYS, "cannot shutdown", 0);
	}
	L (TRA, "closing %d", ctx->poll_arr [i].fd);
	close (ctx->poll_arr [i].fd);
	
	memmove (& ctx->poll_arr [i], & ctx->poll_arr [i+1], (ctx->poll_cnt - i)*sizeof (struct pollfd));
	hcd_client_free (ctx->cli_arr [i]);
	memmove (& ctx->cli_arr [i], & ctx->cli_arr [i+1], (ctx->poll_cnt - i)*sizeof (t_hcd_ctx*));

	ctx->poll_cnt --;

	return (0);

}
int hcd_client__redirect_to_chat (t_hcd_client *client)
{
	hcd_buff__printf (& client->b_out, 
		"HTTP/1.1 301 Moved Permanently\r\n"
		"Cache-Control: no-cache\r\n"
		"Location: /chat/%s\r\n\r\n", client->chat->hsh);
	return (0);
}

int hcd_client__serve_not_found (t_hcd_client *client)
{
	hcd_buff__printf (& client->b_out, 
		"HTTP/1.1 404 Not found\r\n\r\n",0);
	return (0);
}

/* send main (root) page */
int hcd_client__serve_static (t_hcd_ctx *ctx, t_hcd_client *client, t_hcd_request *req)
{
	char *v_a, *v_b;
	t_hcd_chat *c;

	v_a = 
#include "index_a.html.c"
		;
	v_b = 
#include "index_b.html.c"
		;
	hcd_buff__printf (& client->b_out, "HTTP/1.1 200 OK\r\nCache-Control: no-cache\r\nConnection: close\r\n\r\n%s", v_a);

	for (c = ctx->chats; c; c = c->next) {

		/*hcd_buff__printf (& client->b_out, "<li><a href=\"/chat/%s\">%s</a></li>", c->hsh, c->title?c->title:c->hsh);*/
		if (! (c->flags & CHAT_PUBLIC)) continue;

		hcd_buff__printf (& client->b_out, "<li><a href=\"/chat/%s\">", c->hsh);
		hcd_buff__append_esc (& client->b_out, c->title?c->title:c->hsh);
		hcd_buff__printf (& client->b_out, "</a></li>");
	}
	hcd_buff__printf (& client->b_out, "%s", v_b);

	
	return (0);
}

/* send the HTML page of specific chat */
int hcd_client__serve_chat_get (t_hcd_client *client, t_hcd_request *req) 
{
	char *v;

	if (! client->user) {
		client->user = hcd_user__new (client->chat);
	}

	v = 
#include "chat.html.c"
		;

	hcd_buff__printf (& client->b_out, 
			"HTTP/1.1 200 OK\r\n"
			"Content-type: text/html\r\n"
			"Set-Cookie: USER=%s; Path=/send/%s\r\n"
			"Cache-Control: no-cache\r\nConnection: close\r\n\r\n",
			client->user->hsh, client->chat->hsh
			);
	hcd_buff__printf (& client->b_out, v, client->chat->hsh, client->chat->hsh, client->chat->msqu->min?client->chat->msqu->min+1:1);

	return (0);
}

/* send message to client (note the header fields like Time and UserHash) */
void hcd_client__send_msg (t_hcd_client *client, t_hcd_msg *msg)
{
	hcd_buff__printf (& client->b_out, 
			"HTTP/1.1 200 OK\r\n"
			"Content-type: text/html\r\n"
			"Cache-Control: no-cache\r\n"
			"Connection: close\r\n"
			"UserHash: %s\r\n"
			"Time: %ld\r\n\r\n"
			"%s", msg->user->hsh, msg->tim, msg->msg);

	client->b_out.hdstat = 7;
}

int hcd_client__serve_msg (t_hcd_client *client, t_hcd_request *req)
{
	t_hcd_client *c;
	hcd_buff__printf (& client->b_out, "HTTP/1.1 200 OK\r\nCache-Control: no-cache\r\n\r\n");
	L (TRA, "msg=\'%s\'", client->b_in.buff + client->b_in.hdend);
	hcd_msqu__add_msg (client->chat->msqu, client, client->b_in.buff + client->b_in.hdend);

	/* walk through clients of this chat */
	for (c = client->chat->clients; c; c = c->cl_nxt) {
		t_hcd_msg *m;

		m = hcd_msqu__find (c->chat->msqu, c->req->id);

		L (TRA, "CHAT client %p msg %p", c, m);

		if (m) hcd_client__send_msg (c, m);
	}
	return (0);
}

/**
 * Main loop
 */
int ctx__work (t_hcd_ctx *ctx)
{
	int ret, i, r;
	t_hcd_client *cli;

	do {
		/* poll on the array of all opened sockets */
		ret = poll (ctx->poll_arr, ctx->poll_cnt, ctx->interval);

		L (TRA, "poll = %d", ret);

		if (ret < 0) {
			L (ERR|SYS, "poll failed %d", ret);
			return (ret);
		}

		if (ret == 0) continue;

		for (i = 0; i < ctx->poll_max && ret > 0; i ++) {
			if (ctx->poll_arr [i].revents == 0) continue;
			ret --;
			cli = ctx->cli_arr [i];

			L (TRA, "fd %d %d %d", ctx->poll_arr [i].fd, ctx->poll_arr [i].revents, 0);

			/* is it our "accepting" socket ? */
			if (i == 0 && ctx->poll_arr [i].revents == POLLIN) {
				r = ctx__accept (ctx, i);
			} else if (ctx->poll_arr [i].revents & POLLOUT) {
				if (cli->b_out.len - cli->b_out.send_start <= 0) {
					ctx->poll_arr [i].events ^= POLLOUT;

					ctx__shutdown (ctx, i);
					i --;
					continue;
				} else {
					/* socket is free to send */
					r = send (ctx->poll_arr [i].fd, cli->b_out.buff + cli->b_out.send_start,
									cli->b_out.len - cli->b_out.send_start, 0);

					if (r == -1) {
						L (ERR|SYS, "Cannot send buff", 0);
						ctx__shutdown (ctx, i);
						i --;
						continue;
					}
					else cli->b_out.send_start += r;

					if (cli->b_out.send_start == cli->b_out.len) {
						ctx__shutdown (ctx, i);
						i --;
						continue;
					}
				}
			} else {
				char buff [4096];
				/* we can recv from socket */
				while ((r = recv (ctx->poll_arr [i].fd, buff, sizeof (buff), 0)) > 0) {
					L (TRA, "read %d bytes from client id %d: %.*s", r, i, r, buff);

					if (hcd_buff_append (& cli->b_in, buff, r)) {
						L (ERR, "Cannot append", 0);
					}

					/* we've received entire header */
					if (cli->b_in.hdstat == 4) {
						t_hcd_request *req;
						L (TRA, "Header complete...", 0);

						/* let's parse the http header and determine what kind of
						 * request we received */
						if ((req = hcd_client_parse_header (ctx, cli)) == NULL) {
							L (ERR, "Cannot parse header\n", 0);
							hcd_client__serve_not_found (cli);
							ctx->poll_arr [i].events = POLLOUT;
							cli->b_out.send_start = 0;
							r = 1;
							break;
						}

						cli->req = req;
						
						if (req->type == REQ_CHAT_CREATE) {
							t_hcd_chat *chat;

							chat = hcd_chat__new (ctx);
							if (req->flags & F_CHAT_PUBLIC) chat->flags |= CHAT_PUBLIC;
							if (req->data) {
								chat->title = (char*) req->data;
								req->data = NULL;
							}

							cli->chat = chat;

							hcd_client__redirect_to_chat (cli);
							ctx->poll_arr [i].events |= POLLOUT;
							cli->b_out.send_start = 0;
						} else if (req->type == REQ_STATIC_GET) {
							L (TRA, "Get static \'%s\'", req->path?req->path:"<null>");
							hcd_client__serve_static (ctx, cli, req);
							ctx->poll_arr [i].events |= POLLOUT;
							cli->b_out.send_start = 0;

							L (TRA, "Get static \'%s\' donn", req->path?req->path:"<null>");
						} else if (req->type == REQ_CHAT_GET) {
							L (TRA, "Serve chat get \'%s\'", cli->chat->hsh);
							hcd_client__serve_chat_get (cli, req);
							ctx->poll_arr [i].events |= POLLOUT;
							cli->b_out.send_start = 0;
						} else if (req->type == REQ_MSG_POST) {
							/* we have to mark that we're not done yet - we have to
							 * receive the entire message */
							cli->b_in.hdstat = 5;
						} else if (req->type == REQ_MSG_GET) {
							L (TRA, "Client %p requests msg %d", cli, req->id);
							if (req->id < cli->chat->msqu->min) {
								L (TRA, "Client wants id below min (%d, %d)", req->id, cli->chat->msqu->min);
								hcd_client__serve_not_found (cli);
								ctx->poll_arr [i].events = POLLOUT;
								cli->b_out.send_start = 0;
								r = 1;
								break;

							} else if (req->id > cli->chat->msqu->max) {
								L (TRA, "Attaching client %p to chat %s", cli, cli->chat->hsh);
								hcd_chat__attach_client (cli->chat, cli);
								ctx->poll_arr [i].events = 0;
							} else {
								t_hcd_msg *m;

								m = hcd_msqu__find (cli->chat->msqu, cli->req->id);

								L (TRA, "CHAT client %p msg %p %d", cli, m, i);

								if (m) {
									hcd_client__send_msg (cli, m);
								}
							}
						}
					}
				}
				if (r == 0) {
					if (cli->b_in.hdstat == 5) {
						hcd_client__serve_msg (cli, cli->req);
					}
					ctx__shutdown (ctx, i);
					i --;
					continue;
				} 

				L (TRA, "%d %d %d", cli->b_in.hdstat, cli->b_in.explen,cli->b_in.len - cli->b_in.hdend);
					

				if (cli->b_in.hdstat == 5 && cli->b_in.explen == cli->b_in.len - cli->b_in.hdend) {
					hcd_client__serve_msg (cli, cli->req);

					ctx->poll_arr [i].events = POLLOUT;
					cli->b_out.send_start = 0;
				}
			}
		}

		for (i = 1; i < ctx->poll_cnt; i ++) {
			if (ctx->cli_arr [i]->b_out.hdstat == 7) {
				ctx->poll_arr [i].events ^= POLLOUT;
			}
		}
	} while (1);

	return (0);

}

/* continue in detached child */
static void daemonize ()
{
	pid_t p;

	p = fork ();

	if (p < 0) exit (1);
	if (p > 0) exit (0);
	
	setsid ();

	chdir (P_tmpdir);

	signal (SIGCHLD, SIG_IGN);
}

int main (int argc, char **argv)
{
	t_hcd_ctx *ctx;
	int ret, r, c, isdaemon, showhelp;

	isdaemon = 0;
	showhelp = 0;

	srand (time (NULL));
	hcd_set_log_file (stderr);

	ctx = ctx__new (1000);
	ctx->port = DEFAULT_PORT;
	ctx->listcnt = DEFAULT_LISTCNT;
	ctx->interval = -1;

	/* first: let's parse the cmdline arguments */
	while ((c = getopt (argc, argv, "hdp:")) != -1) {
		switch (c) {
			case 'h':
				showhelp = 1;
				break;
			case 'd':
				isdaemon = 1;
				break;
			case 'p':
				ctx->port = atoi (optarg);
				break;
			case '?':
				fprintf (stderr, "Unknown argument '%c'\n", optopt);
				showhelp = 2;
			default:
				fprintf (stderr, "getopt error (%d,%s)\n", optopt, optarg?optarg:"<null>");
				showhelp = 2;
		}
	}

	if (ctx->port == 0) {
		fprintf (stderr, "You must specify valid port number.\n");
		showhelp = 2;
	}

	if (showhelp) {
		fprintf (stderr, "%s [-d] [-h] [-p <portnum>]\n\t-d daemonize\n\t-h show "
				"help\n\t-p port to bind (default %d)\n", argv [0], DEFAULT_PORT);
		exit (showhelp - 1);
	}

	ret = 1;

	do {

		if (isdaemon) daemonize ();

		/* bind & accept main socket */
		r = ctx__bind (ctx);
		if (r) break;

		/* let's do the work */
		r = ctx__work (ctx);
		if (r) break;
			
		ret = 0;

	} while (0);

	if (ret) {
		L (ERR, "error %d occured", ret);
		return (1);
	}

	return (0);
}
