#include <hcd.h>
#include <hcd_reqparse.h>
#include <string.h>
#include <ctype.h>

static int log_level = ~TRA2;
static FILE *log_file = NULL;

void hcd_set_log_file (FILE *f)
{
	log_file = f;
}

int hcd_buff_append (t_hcd_buff *hb, const char *buff, int bsize)
{
	int i;
	L (TRA2, "BUFF_APPEND::BEGIN", 0);
	if (hb->len + bsize > BUFFLEN) {
		L (ERR, "Size %d exceeds BUFFLEN", hb->len + bsize);
		return (1);
	}
	L (TRA2, "BUFF_APPEND::_FT %d %d", hb->len, bsize, 0);
	memcpy (hb->buff + hb->len, buff, bsize);
	L (TRA2, "BUFF_APPEND::AFT", 0);

	if (hb->hdstat < 4) {
		for (i = 0; i < bsize; i ++) {
			char c;
			c = buff [i];
			if (c != '\r' && c != '\n') {
				hb->hdstat = 0;
				continue;
			}
			if (hb->hdstat == 0 || hb->hdstat == 2) {
				if (c == '\r') hb->hdstat ++;
				else hb->hdstat = 0;
			} else if (hb->hdstat == 1 || hb->hdstat == 3) {
				if (c == '\n') hb->hdstat ++;
				else hb->hdstat = 0;
			}
			if (hb->hdstat == 4) {
				hb->hdend = hb->len + i + 1;
				break;
			}
		}
	}
	hb->len += bsize;
	L (TRA2, "BUFF_APPEND::END", 0);
	return (0);
}

t_hcd_request *hcd_request__new ()
{
	return (NEW (t_hcd_request));
}

void hcd_request__free (t_hcd_request *req)
{
	if (req->path) free (req->path);
	if (req->data) free (req->data);
	free (req);
}

char *mstrndup (const char *s, size_t l)
{
	int m;
	char *r;

	for (m = 0; s [m] != '\0' && m < l; m++);

	r = malloc (m + 1);

	memcpy (r, s, m);
	r [m] = '\0';

	return (r);
}
t_hcd_chat *hcd_ctx__chat_find (t_hcd_ctx *ctx, const char *hsh)
{
	t_hcd_chat *ch;

	for (ch = ctx->chats; ch && strcmp (ch->hsh, hsh); ch = ch->next);

	return (ch);
}

/**
 * Parse header and determine type of request 
 */
#ifdef _OLD_REQ_PARSE
t_hcd_request *hcd_client_parse_header (t_hcd_ctx *ctx, t_hcd_client *cl)
{
	t_hcd_buff *b;
	t_hcd_request *req;
	char *p;

	req = hcd_request__new ();
	b = & cl->b_in;
	b->buff [b->len] = '\0';

	if (! strncmp (b->buff, "GET /_", 6) ||
			! strncmp (b->buff, "GET / ", 6)
		) {
		req->type = REQ_STATIC_GET;
		p = strchr (b->buff + 5, ' ');
		if (p) {
			req->path = mstrndup (b->buff + 5, p - (b->buff + 5));
		} else {
			req->path = NULL;
		}
	} else if (! strncmp (b->buff, "GET /chat_create", 16)) {
		req->type = REQ_CHAT_CREATE;

		if (strstr (b->buff, "public=on")) 
			req->flags |= F_CHAT_PRIVATE;
		else
			req->flags = 0;
	} else if (! strncmp (b->buff, "GET /msg/", 9)) {
		t_hcd_chat *chat;
		p = strchr (b->buff + 9, '/');

		if (p) *p = '\0';
		else {
			L (ERR, "Request of incompatible format %s", b->buff + 9);
			goto error;
		} 

		chat = hcd_ctx__chat_find (ctx, b->buff + 9);
		if (! chat) {
			L (ERR, "Request for nonexisting chat %s", b->buff + 9);
			goto error;
		}
		req->id = atoi (p + 1);
		cl->chat = chat;
		req->type = REQ_MSG_GET;

	} else if (! strncmp (b->buff, "GET /chat/", 10)) {
		t_hcd_chat *chat;
		p = strchr (b->buff + 10, ' ');

		if (p) *p = '\0';

		chat = hcd_ctx__chat_find (ctx, b->buff + 10);
		if (! chat) {
			L (ERR, "Request for nonexisting chat %s", b->buff + 10);
			goto error;
		} 
		cl->chat = chat;
		req->type = REQ_CHAT_GET;
	} else if (! strncmp (b->buff, "POST /send/", 11)) {
		t_hcd_chat *chat;
		p = strchr (b->buff + 11, ' ');

		if (p) *p = '\0';

		chat = hcd_ctx__chat_find (ctx, b->buff + 11);
		if (! chat) {
			L (ERR, "Request for nonexisting chat %s", b->buff + 10);
			goto error;
		} 
		cl->chat = chat;

		p = p?p+1:b->buff;
		while ((p = strchr (p+1, '\n'))) {
			L (TRA2, "HD: %.*s", 10, p+1);
			if (! strncmp (p + 1, "Content-Length: ", 16)) {
				b->explen = atoi (p + 1 + 16);
			} else if (! strncmp (p + 1, "Cookie: USER=", 13)) {
				char *n;
				t_hcd_user *user;
				n = strchr (p + 14, '\r');
				L (TRA, "HEME ", 0);
				if (n) {
					*n = '\0';
					user = hcd_chat__user_find (chat, p + 14);
					cl->user = user;
					L (TRA, "User \'%s\' = %p", p + 14, user);
				}
			}
		}

		if (! cl->user) {
			L (ERR, "User not set", 0);
			goto error;
		} 
		req->type = REQ_MSG_POST;
	
	} else {
		L (ERR, "Unknown header (%.*s...)", 10, b->buff);
		goto error;
	}

	return (req);
error:
	hcd_request__free (req);
	return (NULL);
}
#else

t_hcd_request *hcd_client_parse_header (t_hcd_ctx *ctx, t_hcd_client *cl)
{
	t_hcd_request *req;
	struct t_parsectx *pctx;
	t_hcd_buff *b;

	b = & cl->b_in;
	b->buff [b->len] = '\0';

	req = hcd_request__new ();
	pctx = parsectx__new (ctx, cl, req);

	if (parsectx__proceed (pctx, b->buff)) {
		L (ERR, "Cannot parse..", 0);
		goto error;
	}

	if (req->type == 0) {
		L (ERR, "Cannot determine req type", 0);
		goto error;
	}

	if (req->type == REQ_MSG_POST) {
		if (! cl->user) {
			L (ERR, "User not set", 0);
			goto error;
		} 
	}


	parsectx__free (pctx);

	return (req);
error:
	hcd_request__free (req);
	return (NULL);
}


	
#endif

int hcd_buff__printf (t_hcd_buff *buff, const char *fmt, ...)
{
	va_list arg;
	int ret;

	va_start (arg, fmt);
	ret = vsnprintf (buff->buff + buff->len, BUFFLEN - buff->len, fmt, arg);
	buff->len += ret;
	va_end (arg);

	return (ret);
}

int hcd_buff__append_esc (t_hcd_buff *buff, const char *s)
{
	int ret, r, c;
	r = 0;

	while ((c =* s) != 0) {
		if (c == '%' && isxdigit (* (s + 1)) && isxdigit (* (s + 2))) {
			ret = snprintf (buff->buff + buff->len, BUFFLEN - buff->len, "&#x%c%c;", *(s+1), *(s+2));
			s += 3;
		} else if (c == '<') {
			ret = snprintf (buff->buff + buff->len, BUFFLEN - buff->len, "&lt;");
			s += 1;
		} else if (c == '+') {
			ret = snprintf (buff->buff + buff->len, BUFFLEN - buff->len, "&nbsp;");
			s += 1;
		} else {
			ret = snprintf (buff->buff + buff->len, BUFFLEN - buff->len, "%c", c);
			s += 1;
		}
		buff->len += ret;
		r += ret;
	}

	return (ret);
}



void hcd_log (int level, const char *fmt, ...)
{
	va_list arg;
	time_t tim;
	struct tm *tme;

	tim = time (NULL);
	tme = localtime (& tim);

	if (! (level & log_level)) return;

	fprintf (log_file, "%02d:%02d:%02d ", tme->tm_hour, tme->tm_min, tme->tm_sec);

	if (level & ERR) {
		fprintf (log_file, "ERR ");
	} else if (level & TRA) {
		fprintf (log_file, "TRA ");
	} else if (level & SYS) {
		fprintf (log_file, "SYS ");
	} else if (level & DTA) {
		fprintf (log_file, "DTA ");
	} else {
		fprintf (log_file, "MSG ");
	}
	va_start (arg, fmt);

	fseek (log_file, 0L, SEEK_END);

	vfprintf (log_file, fmt, arg);

	va_end (arg);
	
	if (level & SYS) {
		char *s;
		s = strerror (errno);
		if (s) fprintf (log_file, "(%s)", s);
	}

	fprintf (log_file, "\n");

	fflush (log_file);
}

t_hcd_client *hcd_client_new (int fd)
{
	t_hcd_client *cl;

	cl = NEW (t_hcd_client);
	cl->fd = fd;

	return (cl);
}

void hcd_chat__attach_client (t_hcd_chat *chat, t_hcd_client *cl)
{
	if (chat->clients) {
		chat->clients->cl_prev = cl;
		cl->cl_nxt = chat->clients;
		chat->clients = cl;
	} else {
		chat->clients = cl;
		cl->cl_prev = NULL;
		cl->cl_nxt = NULL;
	}
}

void hcd_client_free (t_hcd_client *cl)
{
	if (cl->chat) {
		if (cl->cl_prev) {
			cl->cl_prev->cl_nxt = cl->cl_nxt;
			if (cl->cl_nxt) cl->cl_nxt->cl_prev = cl->cl_prev;
		} else if (cl->chat->clients == cl) {
			cl->chat->clients = cl->cl_nxt;
			if (cl->cl_nxt) cl->cl_nxt = NULL;
		}
	}
	if (cl->req) {
		hcd_request__free (cl->req);
	}
	free (cl);
}

t_hcd_chat *hcd_chat__new (t_hcd_ctx *ctx)
{
	t_hcd_chat *ch;
	int i;

	ch = NEW (t_hcd_chat);

	for (i = 0; i < CHHSHSIZE; i ++) {
		ch->hsh [i] = rand ()%('z' - 'a') + 'a';
	}
	ch->hsh [CHHSHSIZE] = '\0';

	L (TRA, "Creating chat %s", ch->hsh);

	ch->next = ctx->chats;
	ctx->chats = ch;

	ch->msqu = hcd_msqu__new ();

	return (ch);
}

t_hcd_msqu *hcd_msqu__new ()
{
	t_hcd_msqu *q;
	q = NEW (t_hcd_msqu);
	return (q);
}
t_hcd_msg *hcd_msqu__find (t_hcd_msqu *q, int id)
{
	id --;
	if (! (id >= q->min && id <= q->max)) return (NULL);

	return (& q->msgs [id % MSGMAX]);
}

void hcd_msg__set (t_hcd_msg *m, int id, t_hcd_client *cli, const char *msg)
{
	if (m->msg) free (m->msg);
	m->msg = strdup (msg);
	m->cli = cli;
	m->user = cli->user;
}

t_hcd_msg *hcd_msqu__add_msg (t_hcd_msqu *q, t_hcd_client *cli, const char *msg)
{
	t_hcd_msg *g;
	int id;

	id = q->max;

	g = & q->msgs [id % MSGMAX];
	if (id / MSGMAX > 0) q->min ++;
	L (TRA, "Setting msg %d for idx %d", id + 1, id % MSGMAX);

	time (& g->tim);
	hcd_msg__set (g, id + 1, cli, msg);

	q->max ++;

	return (g);
}

void hcd_msqu__free (t_hcd_msqu *q)
{
	
	free (q);
}


t_hcd_user *hcd_user__new (t_hcd_chat *chat)
{
	t_hcd_user *user;
	int i;

	user = NEW (t_hcd_user);
	for (i = 0; i < CHHSHSIZE; i ++) 
		user->hsh [i] = rand ()%('z' - 'a') + 'a';

	user->hsh [CHHSHSIZE] = '\0';

	user->next = chat->users;
	chat->users = user;

	return (user);
}

t_hcd_user *hcd_chat__user_find (t_hcd_chat *chat, const char *hsh)
{
	t_hcd_user *user;

	for (user = chat->users; user && strcmp (user->hsh, hsh); user = user->next) L (TRA, "USer %s != %s;", user->hsh, hsh);

	return (user);
}
